<?php

/**
 * @file
 * Dynamic resources for aggregated CSS and JS files.
 */

/**
 * Aggregator.
 * 
 * @package Aggregation
 */
abstract class AggregationResource extends DynamicResourceFile
{
  /**
   * Array of file to aggregate.
   * 
   * @var array
   */
  protected $_map;

  /**
   * Set files map.
   * 
   * @param array $map
   *   Ordered file list.
   */
  public function setMap(array $map) {
    $this->_map = $map;
  }
}

/**
 * Specific resource implementation for CSS.
 * 
 * This implementation will never use gzip.
 * 
 * @package Aggregation
 */
class AggregationResourceCss extends AggregationResource
{
  /**
   * This is a really specific case, because core already builds the file for
   * us we have to use its own function for this.
   * 
   * @see DynamicResourceFile::build()
   */
  public function build() {
    if (!file_exists($this->_uri)) {
      $this->_checkPath();
      
      $css = array();
  
      // Build some array for core to find out files.
      foreach ($this->_map as $path) {
        $css[] = array('data' => $path, 'type' => 'file');
      }

      if ($coreUri = drupal_build_css_cache($css)) {
        if (!file_unmanaged_copy($coreUri, $this->_uri, FILE_EXISTS_REPLACE)) {
          // Internal server error.
          throw new Exception;
        }

        if ($this->_gzip) {
          $gzipedCoreUri = $coreUri . '.gz';
          $gzipedUri = $this->_uri . '.gz';

          if (file_exists($gzipedCoreUri)) {
            file_unmanaged_copy($gzipedCoreUri, $gzipedUri, FILE_EXISTS_REPLACE);
          }
          else {
            // Core rely on a performance settings that does not really need to
            // exist. Let's just create the file. We don't really care for it
            // to be used or not.
            file_unmanaged_save_data(gzencode(file_get_contents($this->_uri), 9, FORCE_GZIP), $gzipedUri, FILE_EXISTS_REPLACE);
          }
        }
      }
      else {
        // Core was unable to aggregate the files.
        throw new Exception;
      }
    }
  }

  public function getHttpHeaders() {
    return array(
      'Content-Type' => 'text/css',
      'Content-Length' => filesize($this->_uri),
      'Content-Transfer-Encoding' => 'binary',
    );
  }

  /**
   * Void implementation because we are not going to use it.
   * 
   * @see DynamicResourceFile::_buildFileContent()
   */
  public function _buildFileContent() {}
}

/**
 * Specific resource implementation for JS.
 * 
 * FIXME: Implement it. Right now we don't use it at all because core doesn't
 * allow JS aggregation process modification the way he does for CSS. This
 * means that in order to achieve this, we are need to use the hook_js_alter().
 * Using this hook is dangerous because it will force us to give a really high
 * module weight in order to ensure that any other module gets it before us.
 * 
 * @package Aggregation
 */
class AggregationResourceJs extends AggregationResource
{
  public function _buildFileContent() {
    //FIXME: This where we'd want it to be pluggable.
    throw new Exception("Not implemented yet.");
  }

  public function getHttpContentType() {
    return "text/javascript";
  }
}

/**
 * Aggregated files provider.
 * 
 * @package Aggregation
 */
class AggregationResourceProvider extends DynamicResourceProvider
{
  public function getResource($resourcePath, $uri) {

    if (FALSE === strpos($resourcePath, '/')) {
      throw new DynamicResourceNotFoundException;
    }

    list($version, $hash) = explode('/', $resourcePath, 2);

    if (FALSE === strpos($hash, '.')) {
      throw new DynamicResourceNotFoundException;
    }

    list($hash, $type) = explode('.', $hash, 2);

    // Attempt to find existing map.
    $map = db_select('resource_aggregation', 'a')
      ->fields('a', array('map'))
      ->condition('a.hash', $hash)
      ->execute()
      ->fetchField();

    if (!isset($map)) {
      throw new DynamicResourceNotFoundException;
    }

    $map = unserialize($map);

    if (!isset($map) || empty($map)) {
      throw new DynamicResourceNotFoundException;
    }

    switch ($type) {
      case 'js':
        $resource = new AggregationResourceJs($resourcePath, $uri);
        $resource->setMap($map);
        break;

      case 'css':
        $resource = new AggregationResourceCss($resourcePath, $uri);
        $resource->setMap($map);
        break;

      default:
        throw new DynamicResourceNotFoundException;
    }

    return $resource;
  }
}
