<?php

/**
 * @file
 * Dynamic resources for image style.
 */

/**
 * Image styles resource.
 * 
 * @package ImageStyle
 */
class ImageStyleResourceImage extends DynamicResourceFile
{
  /**
   * @var string
   */
  protected $_returnScheme;

  /**
   * Set return scheme.
   * 
   * @param string $scheme
   * 
   * @return ImageStyleResourceImage
   *   Self reference for chaining.
   */
  public function setReturnScheme($scheme) {
    $this->_returnScheme = $scheme;
    return $this;
  }

  /**
   * @var string
   */
  protected $_originalImageUri;

  /**
   * Set original image URI.
   * 
   * @param string $uri
   * 
   * @return ImageStyleResourceImage
   *   Self reference for chaining.
   */
  public function setOriginalImageUri($uri) {
    $this->_originalImageUri = $uri;
    return $this;
  }

  /**
   * @var array
   */
  protected $_style;

  /**
   * Set derivative style.
   * 
   * @param array $style
   * 
   * @return ImageStyleResourceImage
   *   Self reference for chaining.
   */
  public function setStyle($style) {
    $this->_style = $style;
    return $this;
  }

  /**
   * This is a specific use case because the image toolkit will generate the
   * image without giving us the chance to get the output buffer (which is
   * better anyway).
   * 
   * @see DynamicResourceFile::build()
   */
  public function build() {
    if (!file_exists($this->_uri)) {
      $this->_checkPath();
      if (!image_style_create_derivative($this->_style, $this->_originalImageUri, $this->_uri)) {
        throw new Exception("Unable to create derivative image.");
      }
    }
  }

  public function getHttpHeaders() {
    if ($image = image_load($this->_uri)) {
      return array(
        'Content-Type' => $image->info['mime_type'],
        'Content-Length' => $image->info['file_size'],
        'Content-Transfer-Encoding' => 'binary',
      );
    }
    else {
      return parent::getHttpHeaders();
    }
  }

  /**
   * Void implementation because we are not going to use it.
   * 
   * @see DynamicResourceFile::_buildFileContent()
   */
  public function _buildFileContent() {}
}

/**
 * Private files are the only special case because we should definitely not
 * save the thumbnail into a public folder.
 * 
 * FIXME: TODO.
 */
class ImageStyleResourceImagePrivate extends ImageStyleResourceImage {}

/**
 * Image styles resource provider.
 * 
 * @package ImageStyle
 */
class ImageStyleResourceProvider extends DynamicResourceProvider
{
  public function getResource($resourcePath, $uri) {
    list($styleName, $scheme, $filePath) = explode('/', $resourcePath, 3);

    // Do some consistency checks.
    if (!$style = image_style_load($styleName)) {
      throw new DynamicResourceNotFoundException("Unknown file style: '" . $styleName . "'.");
    }
    if (!isset($scheme)) {
      throw new DynamicResourceNotFoundException("Unable to find scheme for asked file: '" . $uri . "'.");
    }

    // Fetch original image URI thanks to given URI.
    $imageUri = $scheme . '://' . $filePath;

    switch ($scheme) {
      case 'private':
        $resource = new ImageStyleResourceImagePrivate($resourcePath, $uri);
        break;

      default:
        $resource = new ImageStyleResourceImage($resourcePath, $uri);
        break;
    }

    $resource
      ->setStyle($style)
      ->setReturnScheme($scheme)
      ->setOriginalImageUri($imageUri);

    return $resource;
  }
}
