<?php

/**
 * @file
 * DynamicResource package.
 */

/**
 * Custom stream wrapper for resource files.
 * 
 * @package DynamicResource
 */
class ResourceStreamWrapper extends DrupalPublicStreamWrapper
{
  /**
   * Implements abstract public function getDirectoryPath()
   */
  public function getDirectoryPath() {
    if (!$path = variable_get('file_resource_path')) {
      return parent::getDirectoryPath() . '/resources';;
    }
    return $path;
  }

  /**
   * Overrides getExternalUrl().
   *
   * Return the HTML URI of a public file.
   */
  function getExternalUrl() {
    $path = str_replace('\\', '/', $this->getTarget());
    return $GLOBALS['base_url'] . '/' . $this->getDirectoryPath() . '/' . drupal_encode_path($path);
  }
}

/**
 * Exception thrown when a provider or a resource is not found.
 * 
 * @package DynamicResource
 */
class DynamicResourceNotFoundException extends Exception {}

/**
 * Exception thrown when a resource is already locked.
 * 
 * @package DynamicResource
 */
class DynamicResourceLockConflictException extends Exception {}

/**
 * Dynamic resource generator.
 * 
 * @package DynamicResource
 */
interface DynamicResourceInterface
{
  /**
   * Unlimited lifetime.
   */
  const LIFETIME_INFINITE = 0;

  /**
   * Get resource raw content.
   * 
   * @return string
   */
  public function build();

  /**
   * Display the content on the output buffer. This should implemented a
   * buffered method if files are big enough.
   */
  public function display();

  /**
   * Get HTTP headers to set for this particular resource.
   * 
   * @return array
   *   Key/value pairs. Key must be valid HTTP headers, and value the full.
   *   You should at least send the Content-Type header to avoid browsers
   *   to attempt displaying your raw data as plain text.
   */
  public function getHttpHeaders();

  /**
   * Get lifetime of this resource, in seconds.
   * 
   * @return int
   */
  public function getLifetime();

  /**
   * Tells if this resource need a mutex for generation.
   * 
   * @return bool
   */
  public function needsLock();

  /**
   * Get URI for this resource.
   */
  public function getUri();

  /**
   * Default constructor.
   * 
   * Provider will give this file its own URI at construct time, once set, this
   * URI remains immutable for this object lifetime.
   * 
   * @param string $resourcePath
   *   Relative to provider resource path.
   * @param string $uri
   *   Full resource URI.
   */
  public function __construct($resourcePath, $uri);
}

/**
 * Base implementation.
 * 
 * @package DynamicResource
 */
abstract class DynamicResource implements DynamicResourceInterface
{
  protected $_resourcePath;

  protected $_uri;

  public function getUri() {
    return $this->_uri;
  }

  /**
   * This default implementation will give the most basic headers you can
   * imagine. You are strongly advised to override it to ensure browsers
   * best behavior.
   * 
   * @see DynamicResourceInterface::getHttpHeaders()
   */
  public function getHttpHeaders() {
    return array(
      'Content-Type' => 'application/octet-stream',
      'Content-Transfer-Encoding' => 'binary',
    );
  }

  public function __construct($resourcePath, $uri) {
    $this->_resourcePath = $resourcePath;
    $this->_uri = $uri;
  }
}

/**
 * Base implementation for statically cached files.
 * 
 * This implementation will fit for small or medium files, but for largest
 * files, you should probably implement yourself a buffered version of it
 * in order to avoid memory capacity exhausted errors.
 * 
 * It handles GZIP compression. While this is not part of the original design,
 * it only will store a gziped copy of the file suffixed with '.gz', such as
 * actual core CSS and JS handling. For this gzip files to be used, you will
 * need your VirtualHost configuration to carry the needed rewrite rules, as
 * you can see for CSS and JS in default core .htaccess file.
 * 
 * @package DynamicResource
 */
abstract class DynamicResourceFile extends DynamicResource
{
  protected $_gzip = FALSE;

  /**
   * Toggle compression.
   * 
   * @param bool $toggle = TRUE
   *   (optional) Set to FALSE will disable compression.
   * 
   * @return bool
   *   TRUE if Gzip can be enabled (depending on the zlib extension presence).
   */
  public function setGzip($toggle = TRUE) {
    $this->_gzip = $toggle && variable_get('clean_url', 0) && extension_loaded('zlib');
  }

  /**
   * Does this resource uses gzip?
   * 
   * @return bool
   */
  public function useGzip() {
    return $this->_gzip;
  }

  /**
   * Simple helper that, using the internal registered URI will create the
   * target directory if not exists.
   * 
   * @throws Exception
   *   If the directory could not have been created.
   */
  protected function _checkPath() {
    $path = dirname($this->_uri);
    if (!is_dir($path)) {
      if (mkdir($path, 0777, TRUE) === FALSE) {
        throw new Exception("Unable to create path '" . $path . "'");
      }
    }
  }

  /**
   * Build the actual file content.
   * 
   * @return string
   */
  protected abstract function _buildFileContent();

  /**
   * Basic implementation that should work for the rest of the world. This
   * implementation triggers a lazzy file build if file does not exist,
   * depending on the this instance URI.
   * 
   * @see DynamicResourceInterface::build()
   */
  public function build() {
    $buffer = NULL;

    if (!file_exists($this->_uri)) {
      $this->_checkPath();
      $buffer = $this->_buildFileContent();
      file_put_contents($this->_uri, $buffer);
    }

    if ($this->_gzip) {
      $gzipedUri = $this->_uri . '.gz';

      if (!file_exists($gzipedUri)) {
        if (!isset($buffer)) {
          $buffer = file_get_contents($this->_uri);
        }

        // We don't care about this operation success. Gzip is only a bonus
        // feature but is not an essential component.
        // FIXME: Does zlib is able to directly encode an existing file?
        file_unmanaged_save_data(gzencode($buffer, 9, FORCE_GZIP), $gzipedUri, FILE_EXISTS_REPLACE);
      }
    }
  }

  /**
   * This default implementation will introspect the current file using the
   * PHP helpers we have. If you can override it easily, do it at least for
   * performances reasons.
   * 
   * @see DynamicResource::getHttpHeaders()
   */
  public function getHttpHeaders() {
    if (file_exists($this->_uri)) {
      return array(
        'Content-Type' => file_get_mimetype($this->_uri),
        'Content-Length' => filesize($this->_uri),
        'Content-Transfer-Encoding' => 'binary',
      );
    }
    else {
      return array(
        'Content-Type' => 'application/octet-stream',
        'Content-Transfer-Encoding' => 'binary',
      );
    }
  }

  /**
   * This should fit for a lot of implementation. We are generating static
   * files here, they are probably not volatile.
   * 
   * @see DynamicResourceInterface::getLifetime()
   */
  public function getLifetime() {
    return DynamicResourceInterface::LIFETIME_INFINITE;
  }

  /**
   * Because we are actually building this over a static file, this method is
   * not business logic anymore, but pure file handling.
   * 
   * @see DynamicResourceInterface::display()
   */
  public function display() {
    echo file_get_contents($this->_uri);
    // FIXME: Is this valuable for common use cases? CSS are really small files
    // and probably doesn't need chunking, for example, while images would
    // probably highly benefit from it.
    //file_transfer($this->_uri, array());
  }

  /**
   * Most implementations of physical static files needs a lock.
   * 
   * @see DynamicResourceInterface::needsLock()
   */
  public function needsLock() {
    return TRUE;
  }

  public function __construct($resourcePath, $uri) {
    parent::__construct($resourcePath, $uri);
    $this->setGzip(TRUE);
  }
}

/**
 * Dynamic resource provider.
 * 
 * @package DynamicResource
 */
interface DynamicResourceProviderInterface
{
  /**
   * Get resource associated to the current partial path.
   * 
   * @param string $resourcePath
   *   Path relative to this provider base path.
   * @param string $uri
   *   Full resource URI provided by the upper layer.
   * 
   * @return DynamicResourceInterface
   * 
   * @throws DynamicResourceNotFoundException
   *   If no resource exists for this path.
   */
  public function getResource($resourcePath, $uri);

  /**
   * Get provider base path.
   * 
   * @return string
   */
  public function getBasePath();

  /**
   * Default constructor.
   * 
   * @param string $basePath
   *   This provider base path.
   */
  public function __construct($basePath);
}

/**
 * Base implementation for providers.
 * 
 * @package DynamicResource
 */
abstract class DynamicResourceProvider implements DynamicResourceProviderInterface
{
  protected $_basePath;

  public function getBasePath() {
    return $this->_basePath;
  }

  public function __construct($uri) {
    $this->_basePath = $uri;
  }
}

/**
 * Entry point for this API.
 * 
 * This class registers all enabled provider and allow menu callback to fetch
 * them magically based on path.
 * 
 * This class is implemented as a singleton.
 * 
 * @package DynamicResource
 */
class DynamicResourceProviderRegistry
{
  /**
   * Get supported schemes.
   * 
   * @return array
   *   Array of schemes.
   */
  public static function getSupportedSchemes() {
    //return array_keys(file_get_stream_wrappers(STREAM_WRAPPERS_LOCAL));

    // FIXME: For security, and before finding out a better algorithm, return
    // an hardcoded array. The commented line upper works fine, it also exports
    // temporary stream wrapper (potential security issue here).
    return array('resource', 'public');
  }

  /**
   * Find scheme by path.
   * 
   * @param string $path
   * 
   * @return string
   *   Scheme.
   * 
   * @throws Exception
   *   If scheme is not found.
   */
  public static function findSchemeByPath($path) {
    foreach (self::getSupportedSchemes() as $scheme) {
      $streamWrapper = file_stream_wrapper_get_instance_by_scheme($scheme);
      if (0 === strpos($path, $streamWrapper->getDirectoryPath())) {
        return $scheme;
      }
    }
  }

  /**
   * Singleton pattern implementation.
   * 
   * @var DynamicResourceProviderRegistry
   */
  private static $__instances = array();

  /**
   * Get singleton instance using the given scheme.
   * 
   * @param string $scheme
   *   Get the instance associated to the given scheme.
   * 
   * @return DynamicResourceProviderRegistry
   * 
   * @throws Exception
   *   If scheme does not exists as a stream wrapper.
   */
  public static function getInstance($scheme) {
    if (!isset(self::$__instances[$scheme])) {
      self::$__instances[$scheme] = new self($scheme);
    }
    return self::$__instances[$scheme];
  }

  /**
   * Get singleton instance using the given full path. This will attempt to
   * guess the scheme by testing the registered stream wrappers.
   *
   * This is a helper function that uses getInstance().
   * 
   * @param string $path
   *   Full query path.
   * 
   * @return DynamicResourceProviderRegistry
   * 
   * @throws Exception
   *   If scheme does not exists as a stream wrapper or if the given path is not
   *   a scheme registered path.
   */
  public static function getInstanceByPath($path) {
    return self::getInstance(self::findSchemeByPath($path));
  }

  /**
   * Associated scheme.
   * 
   * @var string
   */
  protected $_scheme;

  /**
   * Default constructor.
   * Singleton pattern implementation.
   * 
   * @param string $scheme
   *   Set the stream wrapper scheme to associate.
   * 
   * @throws Exception
   *   If scheme does not exists as a stream wrapper.
   */
  private function __construct($scheme) {
    $this->_scheme = $scheme;
    // Enfore the stream wrapper validity check.
    $this->getWorkingPath();
  }

  /**
   * Internal cache.
   * 
   * FIXME: Uses any other existing registry implementation as base instead of
   * this custom code as soon as better helpers in core appears.
   * 
   * @var array
   */
  protected $_cache;

  /**
   * Build internal cache.
   * 
   * FIXME: See $_cache attribute note.
   */
  protected function _buildCache() {
    if (!isset($this->_cache)) {
      $this->_cache = array();
      foreach (module_invoke_all('resource_provider_info') as $path => $class) {
        if (class_exists($class)) {
          $this->_cache[$path] = $class;
        }
        else {
          watchdog('resource', "Defined resource provider class '@class' does not exist, ignoring.", array('@class' => $class), WATCHDOG_ERROR);
        }
      }
      // Hopefully, we are defining a really simple ruleset, for now. This
      // ruleset conflict can be resolved with a simple reverse alphabetical
      // order.
      krsort($this->_cache);
    }
  }

  /**
   * Registry working path. Should probably be relative to files path in most
   * use cases.
   * 
   * @var string
   */
  protected $_workingPath;

  /**
   * Get working path.
   * 
   * @return 
   */
  public function getWorkingPath() {
    if (!isset($this->_workingPath)) {
      $streamWrapper = file_stream_wrapper_get_instance_by_scheme($this->_scheme);
      if (!$streamWrapper instanceof DrupalLocalStreamWrapper) {
        throw new Exception("Scheme '" . $this->_scheme . "' does not exists or does not register a local public stream wrapper class.");
      }
      $this->_workingPath = $streamWrapper->getDirectoryPath();
    }
    return $this->_workingPath;
  }

  /**
   * Get resource provider from the given path.
   * 
   * @param string $path
   *   Path including full file name relative to working path.
   * 
   * @return DynamicResourceProviderInterface
   * 
   * @throws DynamicResourceNotFoundException
   *   If no provider found.
   */
  public function getProvider($path) {
    $this->_buildCache();

    foreach ($this->_cache as $rule => $class) {
      if (0 === strpos($providerPath, $rule)) {
        return new $class($rule);
      }
    }

    throw new DynamicResourceNotFoundException("Provider for path '" . $path . "' not found.");
  }

  /**
   * Get resource from the given path.
   * 
   * @param string $path
   *   Full path including working path or partial working path relative path.
   * 
   * @return DynamicResourceInterface
   * 
   * @throws DynamicResourceNotFoundException
   *   If no provider found.
   */
  public function getResource($path) {
    $this->_buildCache();

    // Remove the current working path from the path.
    if (0 === strpos($path, $this->_workingPath)) {
      $path = substr($path, strlen($this->_workingPath) + 1);
    }

    // Find the right provider to use.
    foreach ($this->_cache as $rule => $class) {
      if (0 === strpos($path, $rule)) {
        $provider = new $class($rule);

        // And use it in order to get the resource if exists.
        return $provider->getResource(substr($path, strlen($rule) + 1), $this->_scheme . '://' . $path);
      }
    }

    throw new DynamicResourceNotFoundException("Provider for path '" . $path . "' not found.");
  }
}
